"""Misc. constants used in the GUI."""
from PySide6.QtCore import QSize


SMALL_BUTTON_WIDTH = 24
SMALL_BUTTON_HEIGHT = 20
SMALL_BUTTON_SIZE = QSize(
    SMALL_BUTTON_WIDTH,
    SMALL_BUTTON_HEIGHT,
)

SMALL_BUTTON_ICON_WIDTH = 15
SMALL_BUTTON_ICON_HEIGHT = 15
SMALL_BUTTON_ICON_SIZE = QSize(
    SMALL_BUTTON_ICON_WIDTH,
    SMALL_BUTTON_ICON_HEIGHT,
)

BIG_BUTTON_WIDTH = 120
BIG_BUTTON_HEIGHT = 60
BIG_BUTTON_SIZE = QSize(
    BIG_BUTTON_WIDTH,
    BIG_BUTTON_HEIGHT,
)

BIG_BUTTON_ICON_WIDTH = 32
BIG_BUTTON_ICON_HEIGHT = 32
BIG_BUTTON_ICON_SIZE = QSize(
    BIG_BUTTON_ICON_WIDTH,
    BIG_BUTTON_ICON_HEIGHT,
)

TAB_WIDGET_CORNER_BUTTON_WIDTH = 40
TAB_WIDGET_CORNER_BUTTON_HEIGHT = 20
TAB_WIDGET_CORNER_BUTTON_SIZE = QSize(
    TAB_WIDGET_CORNER_BUTTON_WIDTH,
    TAB_WIDGET_CORNER_BUTTON_HEIGHT,
)
